@@ -13,8 +13,18 @@ void CommandData::Init()
 {
   RAROptions::Init();
 
-  *Command=0;
-  *ArcName=0;
+  Command.clear();
+  ArcName.clear();
+  ExtrPath.clear();
+  TempPath.clear();
+  SFXModule.clear();
+  CommentFile.clear();
+  ArcPath.clear();
+  ExclArcPath.clear();
+  LogName.clear();
+  EmailTo.clear();
+  UseStdin.clear();
+
   FileLists=false;
   NoMoreSwitches=false;
 
@@ -26,61 +36,52 @@ void CommandData::Init()
   FileArgs.Reset();
   ExclArgs.Reset();
   InclArgs.Reset();
-  StoreArgs.Reset();
   ArcNames.Reset();
-  NextVolSizes.Reset();
-}
-
-
-// Return the pointer to next position in the string and store dynamically
-// allocated command line parameter in Par.
-static const wchar *AllocCmdParam(const wchar *CmdLine,wchar **Par)
-{
-  const wchar *NextCmd=GetCmdParam(CmdLine,NULL,0);
-  if (NextCmd==NULL)
-    return NULL;
-  size_t ParSize=NextCmd-CmdLine+2; // Parameter size including the trailing zero.
-  *Par=(wchar *)malloc(ParSize*sizeof(wchar));
-  if (*Par==NULL)
-    return NULL;
-  return GetCmdParam(CmdLine,*Par,ParSize);
+  StoreArgs.Reset();
+#ifdef PROPAGATE_MOTW
+  MotwList.Reset();
+#endif
+  Password.Clean();
+  NextVolSizes.clear();
+#ifdef RARDLL
+  DllDestName.clear();
+#endif
 }
 
 
 #if !defined(SFX_MODULE)
 void CommandData::ParseCommandLine(bool Preprocess,int argc, char *argv[])
 {
-  *Command=0;
+  Command.clear();
   NoMoreSwitches=false;
 #ifdef CUSTOM_CMDLINE_PARSER
   // In Windows we may prefer to implement our own command line parser
   // to avoid replacing \" by " in standard parser. Such replacing corrupts
   // destination paths like "dest path\" in extraction commands.
-  // Also our own parser is Unicode compatible.
-  const wchar *CmdLine=GetCommandLine();
+  std::wstring CmdLine=GetCommandLine();
+
+  std::wstring Param;
+  std::wstring::size_type Pos=0;
 
-  wchar *Par;
   for (bool FirstParam=true;;FirstParam=false)
   {
-    if ((CmdLine=AllocCmdParam(CmdLine,&Par))==NULL)
+    if (!GetCmdParam(CmdLine,Pos,Param))
       break;
     if (!FirstParam) // First parameter is the executable name.
       if (Preprocess)
-        PreprocessArg(Par);
+        PreprocessArg(Param.c_str());
       else
-        ParseArg(Par);
-    free(Par);
+        ParseArg(Param.c_str());
   }
 #else
-  Array<wchar> Arg;
   for (int I=1;I<argc;I++)
   {
-    Arg.Alloc(strlen(argv[I])+1);
-    CharToWide(argv[I],&Arg[0],Arg.Size());
+    std::wstring Arg;
+    CharToWide(argv[I],Arg);
     if (Preprocess)
-      PreprocessArg(&Arg[0]);
+      PreprocessArg(Arg.c_str());
     else
-      ParseArg(&Arg[0]);
+      ParseArg(Arg.c_str());
   }
 #endif
   if (!Preprocess)
@@ -90,7 +91,7 @@ void CommandData::ParseCommandLine(bool Preprocess,int argc, char *argv[])
 
 
 #if !defined(SFX_MODULE)
-void CommandData::ParseArg(wchar *Arg)
+void CommandData::ParseArg(const wchar *Arg)
 {
   if (IsSwitch(*Arg) && !NoMoreSwitches)
     if (Arg[1]=='-' && Arg[2]==0)
@@ -98,33 +99,52 @@ void CommandData::ParseArg(wchar *Arg)
     else
       ProcessSwitch(Arg+1);
   else
-    if (*Command==0)
+    if (Command.empty())
     {
-      wcsncpyz(Command,Arg,ASIZE(Command));
+      Command=Arg;
 
 
-      *Command=toupperw(*Command);
+      Command[0]=toupperw(Command[0]);
       // 'I' and 'S' commands can contain case sensitive strings after
       // the first character, so we must not modify their case.
       // 'S' can contain SFX name, which case is important in Unix.
-      if (*Command!='I' && *Command!='S')
+      if (Command[0]!='I' && Command[0]!='S')
         wcsupper(Command);
+      if (Command[0]=='P') // Enforce -idq for print command.
+      {
+        MsgStream=MSG_ERRONLY;
+        SetConsoleMsgStream(MSG_ERRONLY);
+      }
     }
     else
-      if (*ArcName==0)
-        wcsncpyz(ArcName,Arg,ASIZE(ArcName));
+      if (ArcName.empty())
+        ArcName=Arg;
       else
       {
         // Check if last character is the path separator.
         size_t Length=wcslen(Arg);
         wchar EndChar=Length==0 ? 0:Arg[Length-1];
-        bool EndSeparator=IsDriveDiv(EndChar) || IsPathDiv(EndChar);
+        // Check if trailing path separator like path\ is present.
+        bool FolderArg=IsDriveDiv(EndChar) || IsPathDiv(EndChar);
 
-        wchar CmdChar=toupperw(*Command);
+        // 2024.01.05: We were asked to support exotic d:. and d:.. paths.
+        if (IsDriveLetter(Arg) && Arg[2]=='.' && (Arg[3]==0 || Arg[3]=='.' && Arg[4]==0))
+          FolderArg=true;
+
+        // 2024.01.06: FindFile::FastFind check below fails in Windows 10 if
+        // "." or ".." points at disk root. So we enforce it for "." and ".."
+        // optionally preceded with some path like "..\..".
+        size_t L=Length;
+        if (L>0 && Arg[L-1]=='.' && (L==1 || L>=2 && (IsPathDiv(Arg[L-2]) ||
+            Arg[L-2]=='.' && (L==2 || L>=3 && IsPathDiv(Arg[L-3])))))
+          FolderArg=true;
+
+        wchar CmdChar=toupperw(Command[0]);
         bool Add=wcschr(L"AFUM",CmdChar)!=NULL;
         bool Extract=CmdChar=='X' || CmdChar=='E';
-        if (EndSeparator && !Add)
-          wcsncpyz(ExtrPath,Arg,ASIZE(ExtrPath));
+        bool Repair=CmdChar=='R' && Command[1]==0;
+        if (FolderArg && !Add)
+          ExtrPath=Arg;
         else
           if ((Add || CmdChar=='T') && (*Arg!='@' || ListMode==RCLM_REJECT_LISTS))
             FileArgs.AddString(Arg);
@@ -140,11 +160,11 @@ void CommandData::ParseArg(wchar *Arg)
               ReadTextFile(Arg+1,&FileArgs,false,true,FilelistCharset,true,true,true);
 
             }
-            else
-              if (Found && FileData.IsDir && Extract && *ExtrPath==0)
+            else // We use 'destpath\' when extracting and reparing.
+              if (Found && FileData.IsDir && (Extract || Repair) && ExtrPath.empty())
               {
-                wcsncpyz(ExtrPath,Arg,ASIZE(ExtrPath));
-                AddEndSlash(ExtrPath,ASIZE(ExtrPath));
+                ExtrPath=Arg;
+                AddEndSlash(ExtrPath);
               }
               else
                 FileArgs.AddString(Arg);
@@ -172,12 +192,12 @@ void CommandData::ParseDone()
 #if !defined(SFX_MODULE)
 void CommandData::ParseEnvVar()
 {
-  char *EnvStr=getenv("RAR");
-  if (EnvStr!=NULL)
+  char *EnvVar=getenv("RARINISWITCHES");
+  if (EnvVar!=NULL)
   {
-    Array<wchar> EnvStrW(strlen(EnvStr)+1);
-    CharToWide(EnvStr,&EnvStrW[0],EnvStrW.Size());
-    ProcessSwitchesString(&EnvStrW[0]);
+    std::wstring EnvStr;
+    CharToWide(EnvVar,EnvStr);
+    ProcessSwitchesString(EnvStr);
   }
 }
 #endif
@@ -186,7 +206,7 @@ void CommandData::ParseEnvVar()
 
 #if !defined(SFX_MODULE)
 // Preprocess those parameters, which must be processed before the rest of
-// command line. Return 'false' to stop further processing.
+// command line.
 void CommandData::PreprocessArg(const wchar *Arg)
 {
   if (IsSwitch(Arg[0]) && !NoMoreSwitches)
@@ -195,7 +215,7 @@ void CommandData::PreprocessArg(const wchar *Arg)
     if (Arg[0]=='-' && Arg[1]==0) // Switch "--".
       NoMoreSwitches=true;
     if (wcsicomp(Arg,L"cfg-")==0)
-      ConfigDisabled=true;
+      ProcessSwitch(Arg);
     if (wcsnicomp(Arg,L"ilog",4)==0)
     {
       // Ensure that correct log file name is already set
@@ -207,13 +227,13 @@ void CommandData::PreprocessArg(const wchar *Arg)
     {
       // Process -sc before reading any file lists.
       ProcessSwitch(Arg);
-      if (*LogName!=0)
+      if (!LogName.empty())
         InitLogOptions(LogName,ErrlogCharset);
     }
   }
   else
-    if (*Command==0)
-      wcsncpy(Command,Arg,ASIZE(Command)); // Need for rar.ini.
+    if (Command.empty())
+      Command=Arg; // Need for rar.ini.
 }
 #endif
 
@@ -231,10 +251,10 @@ void CommandData::ReadConfig()
         Str++;
       if (wcsnicomp(Str,L"switches=",9)==0)
         ProcessSwitchesString(Str+9);
-      if (*Command!=0)
+      if (!Command.empty())
       {
         wchar Cmd[16];
-        wcsncpyz(Cmd,Command,ASIZE(Cmd));
+        wcsncpyz(Cmd,Command.c_str(),ASIZE(Cmd));
         wchar C0=toupperw(Cmd[0]);
         wchar C1=toupperw(Cmd[1]);
         if (C0=='I' || C0=='L' || C0=='M' || C0=='S' || C0=='V')
@@ -254,14 +274,19 @@ void CommandData::ReadConfig()
 
 
 #if !defined(SFX_MODULE)
-void CommandData::ProcessSwitchesString(const wchar *Str)
+void CommandData::ProcessSwitchesString(const std::wstring &Str)
 {
-  wchar *Par;
-  while ((Str=AllocCmdParam(Str,&Par))!=NULL)
+  std::wstring Par;
+  std::wstring::size_type Pos=0;
+  while (GetCmdParam(Str,Pos,Par))
   {
-    if (IsSwitch(*Par))
-      ProcessSwitch(Par+1);
-    free(Par);
+    if (IsSwitch(Par[0]))
+      ProcessSwitch(&Par[1]);
+    else
+    {
+      mprintf(St(MSwSyntaxError),Par.c_str());
+      ErrHandler.Exit(RARX_USERERROR);
+    }
   }
 }
 #endif
@@ -271,6 +296,9 @@ void CommandData::ProcessSwitchesString(const wchar *Str)
 void CommandData::ProcessSwitch(const wchar *Switch)
 {
 
+  if (LargePageAlloc::ProcessSwitch(this,Switch))
+    return;
+
   switch(toupperw(Switch[0]))
   {
     case '@':
@@ -287,29 +315,49 @@ void CommandData::ProcessSwitch(const wchar *Switch)
             AppendArcNameToPath=APPENDARCNAME_DESTPATH;
           else
             if (Switch[2]=='1')
-              AppendArcNameToPath=APPENDARCNAME_OWNDIR;
+              AppendArcNameToPath=APPENDARCNAME_OWNSUBDIR;
+            else
+              if (Switch[2]=='2')
+                AppendArcNameToPath=APPENDARCNAME_OWNDIR;
           break;
 #ifndef SFX_MODULE
         case 'G':
           if (Switch[2]=='-' && Switch[3]==0)
             GenerateArcName=0;
           else
-          {
-            GenerateArcName=true;
-            wcsncpyz(GenerateMask,Switch+2,ASIZE(GenerateMask));
-          }
+            if (toupperw(Switch[2])=='F')
+              wcsncpyz(DefGenerateMask,Switch+3,ASIZE(DefGenerateMask));
+            else
+            {
+              GenerateArcName=true;
+              wcsncpyz(GenerateMask,Switch+2,ASIZE(GenerateMask));
+            }
           break;
 #endif
         case 'I':
           IgnoreGeneralAttr=true;
           break;
-        case 'N': // Reserved for archive name.
+        case 'M':
+          switch(toupperw(Switch[2]))
+          {
+            case 0:
+            case 'S':
+              ArcMetadata=ARCMETA_SAVE;
+              break;
+            case 'R':
+              ArcMetadata=ARCMETA_RESTORE;
+              break;
+            default:
+              BadSwitch(Switch);
+              break;
+          }
           break;
         case 'O':
           AddArcOnly=true;
           break;
         case 'P':
-          wcsncpyz(ArcPath,Switch+2,ASIZE(ArcPath));
+          // Convert slashes here than before every comparison.
+          SlashToNative(Switch+2,ArcPath);
           break;
         case 'S':
           SyncFiles=true;
@@ -320,7 +368,14 @@ void CommandData::ProcessSwitch(const wchar *Switch)
       }
       break;
     case 'C':
-      if (Switch[2]==0)
+      if (Switch[2]!=0)
+      {
+          if (wcsicomp(Switch+1,L"FG-")==0)
+            ConfigDisabled=true;
+          else
+            BadSwitch(Switch);
+      }
+      else
         switch(toupperw(Switch[1]))
         {
           case '-':
@@ -332,10 +387,15 @@ void CommandData::ProcessSwitch(const wchar *Switch)
           case 'L':
             ConvertNames=NAMES_LOWERCASE;
             break;
+          default:
+            BadSwitch(Switch);
+            break;
         }
       break;
     case 'D':
-      if (Switch[2]==0)
+      if (Switch[2]!=0)
+        BadSwitch(Switch);
+      else
         switch(toupperw(Switch[1]))
         {
           case 'S':
@@ -347,6 +407,9 @@ void CommandData::ProcessSwitch(const wchar *Switch)
           case 'F':
             DeleteFiles=true;
             break;
+          default:
+            BadSwitch(Switch);
+            break;
         }
       break;
     case 'E':
@@ -367,16 +430,23 @@ void CommandData::ProcessSwitch(const wchar *Switch)
             case '3':
               ExclPath=EXCL_ABSPATH;
               break;
+            case '4':
+              // Convert slashes here than before every comparison.
+              SlashToNative(Switch+3,ExclArcPath);
+              break;
+            default:
+              BadSwitch(Switch);
+              break;
           }
           break;
         default:
           if (Switch[1]=='+')
           {
-            InclFileAttr|=GetExclAttr(Switch+2);
+            InclFileAttr|=GetExclAttr(Switch+2,InclDir);
             InclAttrSet=true;
           }
           else
-            ExclFileAttr|=GetExclAttr(Switch+1);
+            ExclFileAttr|=GetExclAttr(Switch+1,ExclDir);
           break;
       }
       break;
@@ -393,13 +463,17 @@ void CommandData::ProcessSwitch(const wchar *Switch)
           EncryptHeaders=true;
           if (Switch[2]!=0)
           {
+            // We use this code for other archive formats too, so MAXPASSWORD
+            // instead of MAXPASSWORD_RAR.
+            if (wcslen(Switch+2)>=MAXPASSWORD)
+              uiMsg(UIERROR_TRUNCPSW,MAXPASSWORD-1);
             Password.Set(Switch+2);
             cleandata((void *)Switch,wcslen(Switch)*sizeof(Switch[0]));
           }
           else
             if (!Password.IsSet())
             {
-              uiGetPassword(UIPASSWORD_GLOBAL,NULL,&Password);
+              uiGetPassword(UIPASSWORD_GLOBAL,L"",&Password,NULL);
               eprintf(L"\n");
             }
           break;
@@ -411,7 +485,7 @@ void CommandData::ProcessSwitch(const wchar *Switch)
     case 'I':
       if (wcsnicomp(Switch+1,L"LOG",3)==0)
       {
-        wcsncpyz(LogName,Switch[4]!=0 ? Switch+4:DefLogName,ASIZE(LogName));
+        LogName=Switch[4]!=0 ? Switch+4:DefLogName;
         break;
       }
       if (wcsnicomp(Switch+1,L"SND",3)==0)
@@ -429,12 +503,12 @@ void CommandData::ProcessSwitch(const wchar *Switch)
       }
       if (wcsnicomp(Switch+1,L"EML",3)==0)
       {
-        wcsncpyz(EmailTo,Switch[4]!=0 ? Switch+4:L"@",ASIZE(EmailTo));
+        EmailTo=Switch[4]!=0 ? Switch+4:L"@";
         break;
       }
-      if (wcsicomp(Switch+1,L"M")==0)
+      if (wcsicomp(Switch+1,L"M")==0) // For compatibility with pre-WinRAR 6.0 -im syntax. Replaced with -idv.
       {
-        MoreInfo=true;
+        VerboseOutput=true;
         break;
       }
       if (wcsicomp(Switch+1,L"NUL")==0)
@@ -461,6 +535,12 @@ void CommandData::ProcessSwitch(const wchar *Switch)
             case 'P':
               DisablePercentage=true;
               break;
+            case 'N':
+              DisableNames=true;
+              break;
+            case 'V':
+              VerboseOutput=true;
+              break;
           }
         break;
       }
@@ -530,13 +610,14 @@ void CommandData::ProcessSwitch(const wchar *Switch)
                 }
                 switch(toupperw(*(Str++)))
                 {
-                  case 'T': Type=FILTER_PPM;         break;
+//                  case 'T': Type=FILTER_TEXT;        break;
                   case 'E': Type=FILTER_E8;          break;
                   case 'D': Type=FILTER_DELTA;       break;
-                  case 'A': Type=FILTER_AUDIO;       break;
-                  case 'C': Type=FILTER_RGB;         break;
-                  case 'I': Type=FILTER_ITANIUM;     break;
-                  case 'R': Type=FILTER_ARM;         break;
+//                  case 'A': Type=FILTER_AUDIO;       break;
+//                  case 'C': Type=FILTER_RGB;         break;
+//                  case 'R': Type=FILTER_ARM;         break;
+                  case 'L': Type=FILTER_LONGRANGE;   break;
+                  case 'X': Type=FILTER_EXHAUSTIVE;  break;
                 }
                 if (*Str=='+' || *Str=='-')
                   State=*(Str++)=='+' ? FILTER_FORCE:FILTER_DISABLE;
@@ -546,42 +627,74 @@ void CommandData::ProcessSwitch(const wchar *Switch)
               }
             }
           break;
-        case 'M':
-          break;
         case 'D':
-          break;
-        case 'S':
           {
-            wchar StoreNames[1024];
-            wcsncpyz(StoreNames,(Switch[2]==0 ? DefaultStoreList:Switch+2),ASIZE(StoreNames));
-            wchar *Names=StoreNames;
-            while (*Names!=0)
+            bool SetDictLimit=toupperw(Switch[2])=='X';
+
+            uint64 Size=atoiw(Switch+(SetDictLimit ? 3 : 2));
+            wchar LastChar=toupperw(Switch[wcslen(Switch)-1]);
+            if (IsDigit(LastChar))
+              LastChar=SetDictLimit ? 'G':'M'; // Treat -md128 as -md128m and -mdx32 as -mdx32g.
+            switch(LastChar)
             {
-              wchar *End=wcschr(Names,';');
-              if (End!=NULL)
-                *End=0;
-              if (*Names=='.')
-                Names++;
-              wchar Mask[NM];
-              if (wcspbrk(Names,L"*?.")==NULL)
-                swprintf(Mask,ASIZE(Mask),L"*.%ls",Names);
-              else
-                wcsncpyz(Mask,Names,ASIZE(Mask));
-              StoreArgs.AddString(Mask);
-              if (End==NULL)
+              case 'K':
+                Size*=1024;
                 break;
-              Names=End+1;
+              case 'M':
+                Size*=1024*1024;
+                break;
+              case 'G':
+                Size*=1024*1024*1024;
+                break;
+              default:
+                BadSwitch(Switch);
+            }
+
+            // 2023.07.22: For 4 GB and less we also check that it is power of 2,
+            // so archives are compatible with RAR 5.0+.
+            // We allow Size>PACK_MAX_DICT here, so we can use -md[x] to unpack
+            // archives created by future versions with higher PACK_MAX_DICT
+            uint Flags;
+            if ((Size=Archive::GetWinSize(Size,Flags))==0 ||
+                Size<=0x100000000ULL && !IsPow2(Size))
+              BadSwitch(Switch);
+            else
+              if (SetDictLimit)
+                WinSizeLimit=Size;
+              else
+              {
+                WinSize=Size;
+              }
+          }
+          break;
+        case 'E':
+          if (toupperw(Switch[2])=='S' && Switch[3]==0)
+            SkipEncrypted=true;
+          break;
+        case 'L':
+          if (toupperw(Switch[2])=='P')
+          {
+            UseLargePages=true;
+            if (!LargePageAlloc::IsPrivilegeAssigned() && LargePageAlloc::AssignConfirmation())
+            {
+              LargePageAlloc::AssignPrivilege();
+
+              // Quit immediately. We do not want to interrupt the current copy
+              // archive processing with reboot after assigning privilege.
+              SetupComplete=true;
             }
           }
           break;
+        case 'M':
+          break;
+        case 'S':
+          GetBriefMaskList(Switch[2]==0 ? DefaultStoreList:Switch+2,StoreArgs);
+          break;
 #ifdef RAR_SMP
         case 'T':
           Threads=atoiw(Switch+2);
           if (Threads>MaxPoolThreads || Threads<1)
             BadSwitch(Switch);
-          else
-          {
-          }
           break;
 #endif
         default:
@@ -627,8 +740,31 @@ void CommandData::ProcessSwitch(const wchar *Switch)
 #ifdef SAVE_LINKS
         case 'L':
           SaveSymLinks=true;
-          if (toupperw(Switch[2])=='A')
-            AbsoluteLinks=true;
+          for (uint I=2;Switch[I]!=0;I++)
+            switch(toupperw(Switch[I]))
+            {
+              case 'A':
+                AbsoluteLinks=true;
+                break;
+              case '-':
+                SkipSymLinks=true;
+                break;
+              default:
+                BadSwitch(Switch);
+                break;
+            }
+          break;
+#endif
+#ifdef PROPAGATE_MOTW
+        case 'M':
+          {
+            MotwAllFields=Switch[2]=='1';
+            const wchar *Sep=wcschr(Switch+2,'=');
+            if (Switch[2]=='-')
+              MotwList.Reset();
+            else
+              GetBriefMaskList(Sep==nullptr ? L"*":Sep+1,MotwList);
+          }
           break;
 #endif
 #ifdef _WIN_ALL
@@ -637,6 +773,10 @@ void CommandData::ProcessSwitch(const wchar *Switch)
             AllowIncompatNames=true;
           break;
 #endif
+        case 'P':
+          ExtrPath=Switch+2;
+          AddEndSlash(ExtrPath);
+          break;
         case 'R':
           Overwrite=OVERWRITE_AUTORENAME;
           break;
@@ -656,11 +796,13 @@ void CommandData::ProcessSwitch(const wchar *Switch)
     case 'P':
       if (Switch[1]==0)
       {
-        uiGetPassword(UIPASSWORD_GLOBAL,NULL,&Password);
+        uiGetPassword(UIPASSWORD_GLOBAL,L"",&Password,NULL);
         eprintf(L"\n");
       }
       else
       {
+        if (wcslen(Switch+1)>=MAXPASSWORD)
+          uiMsg(UIERROR_TRUNCPSW,MAXPASSWORD-1);
         Password.Set(Switch+1);
         cleandata((void *)Switch,wcslen(Switch)*sizeof(Switch[0]));
       }
@@ -727,7 +869,52 @@ void CommandData::ProcessSwitch(const wchar *Switch)
         switch(toupperw(Switch[1]))
         {
           case 0:
+          case '=':
             Solid|=SOLID_NORMAL;
+            if (Switch[1]=='=')
+            {
+              uint Par=0;
+              for (const wchar *S=Switch+2;*S!=0;S++)
+              {
+                if (IsDigit(*S))
+                  Par=Par*10+*S-'0';
+                switch(toupperw(*S))
+                {
+                  case '-':
+                    Solid=SOLID_NONE;
+                    break;
+                  case 'D':
+                    Solid|=SOLID_VOLUME_DEPENDENT;
+                    break;
+                  case 'E':
+                    Solid|=SOLID_FILEEXT;
+                    break;
+                  case 'F':
+                    Solid|=SOLID_COUNT;
+                    SolidCount=Par;
+                    break;
+                  case 'K':
+                    Solid|=SOLID_BLOCK_SIZE;
+                    SolidBlockSize=Par*1024LL;
+                    break;
+                  case 'M':
+                    Solid|=SOLID_BLOCK_SIZE;
+                    SolidBlockSize=Par*1024LL*1024LL;
+                    break;
+                  case 'G':
+                    Solid|=SOLID_BLOCK_SIZE;
+                    SolidBlockSize=Par*1024LL*1024LL*1024LL;
+                    break;
+                  case 'R':
+                    Solid=SOLID_RESET;
+                    break;
+                  case 'V':
+                    Solid|=SOLID_VOLUME_INDEPENDENT;
+                    break;
+                }
+
+              }
+            }
             break;
           case '-':
             Solid=SOLID_NONE;
@@ -741,13 +928,22 @@ void CommandData::ProcessSwitch(const wchar *Switch)
           case 'D':
             Solid|=SOLID_VOLUME_DEPENDENT;
             break;
+          case 'I':
+            ProhibitConsoleInput();
+            // We do not assign the archive name automatically for -si
+            // if archive name is omitted and require the archive name to
+            // present always. Otherwise for"type arc.rar|rar x -si arc2.rar"
+            // if arc2.rar is a dummy archive name or file inside of arc.rar,
+            // which needs to be extracted.
+            UseStdin=Switch[2] ? Switch+2:L"stdin";
+            break;
           case 'L':
             if (IsDigit(Switch[2]))
-              FileSizeLess=atoilw(Switch+2);
+              FileSizeLess=GetModSize(Switch+2,1);
             break;
           case 'M':
             if (IsDigit(Switch[2]))
-              FileSizeMore=atoilw(Switch+2);
+              FileSizeMore=GetModSize(Switch+2,1);
             break;
           case 'C':
             {
@@ -805,12 +1001,6 @@ void CommandData::ProcessSwitch(const wchar *Switch)
     case 'T':
       switch(toupperw(Switch[1]))
       {
-        case 'K':
-          ArcTime=ARCTIME_KEEP;
-          break;
-        case 'L':
-          ArcTime=ARCTIME_LATEST;
-          break;
         case 'O':
           SetTimeFilters(Switch+2,true,true);
           break;
@@ -824,39 +1014,7 @@ void CommandData::ProcessSwitch(const wchar *Switch)
           SetTimeFilters(Switch+2,false,false);
           break;
         case 'S':
-          {
-            EXTTIME_MODE Mode=EXTTIME_HIGH3;
-            bool CommonMode=Switch[2]>='0' && Switch[2]<='4';
-            if (CommonMode)
-              Mode=(EXTTIME_MODE)(Switch[2]-'0');
-            if (Mode==EXTTIME_HIGH1 || Mode==EXTTIME_HIGH2) // '2' and '3' not supported anymore.
-              Mode=EXTTIME_HIGH3;
-            if (Switch[2]=='-')
-              Mode=EXTTIME_NONE;
-            if (CommonMode || Switch[2]=='-' || Switch[2]=='+' || Switch[2]==0)
-              xmtime=xctime=xatime=Mode;
-            else
-            {
-              if (Switch[3]>='0' && Switch[3]<='4')
-                Mode=(EXTTIME_MODE)(Switch[3]-'0');
-              if (Mode==EXTTIME_HIGH1 || Mode==EXTTIME_HIGH2) // '2' and '3' not supported anymore.
-                Mode=EXTTIME_HIGH3;
-              if (Switch[3]=='-')
-                Mode=EXTTIME_NONE;
-              switch(toupperw(Switch[2]))
-              {
-                case 'M':
-                  xmtime=Mode;
-                  break;
-                case 'C':
-                  xctime=Mode;
-                  break;
-                case 'A':
-                  xatime=Mode;
-                  break;
-              }
-            }
-          }
+          SetStoreTimeMode(Switch+2);
           break;
         case '-':
           Test=false;
@@ -894,8 +1052,8 @@ void CommandData::ProcessSwitch(const wchar *Switch)
       }
       break;
     case 'W':
-      wcsncpyz(TempPath,Switch+1,ASIZE(TempPath));
-      AddEndSlash(TempPath,ASIZE(TempPath));
+      TempPath=Switch+1;
+      AddEndSlash(TempPath);
       break;
     case 'Y':
       AllYes=true;
@@ -904,10 +1062,10 @@ void CommandData::ProcessSwitch(const wchar *Switch)
       if (Switch[1]==0)
       {
         // If comment file is not specified, we read data from stdin.
-        wcsncpyz(CommentFile,L"stdin",ASIZE(CommentFile));
+        CommentFile=L"stdin";
       }
       else
-        wcsncpyz(CommentFile,Switch+1,ASIZE(CommentFile));
+        CommentFile=Switch+1;
       break;
     case '?' :
       OutHelp(RARX_SUCCESS);
@@ -934,32 +1092,42 @@ void CommandData::ProcessCommand()
 #ifndef SFX_MODULE
 
   const wchar *SingleCharCommands=L"FUADPXETK";
-  if (Command[0]!=0 && Command[1]!=0 && wcschr(SingleCharCommands,Command[0])!=NULL || *ArcName==0)
-    OutHelp(*Command==0 ? RARX_SUCCESS:RARX_USERERROR); // Return 'success' for 'rar' without parameters.
 
-  const wchar *ArcExt=GetExt(ArcName);
+  // RAR -mlp command is the legitimate way to assign the required privilege.
+  if (Command.empty() && UseLargePages || SetupComplete)
+    return;
+
+  if (Command[0]!=0 && Command[1]!=0 && wcschr(SingleCharCommands,Command[0])!=NULL || ArcName.empty())
+    OutHelp(Command.empty() ? RARX_SUCCESS:RARX_USERERROR); // Return 'success' for 'rar' without parameters.
+
+  size_t ExtPos=GetExtPos(ArcName);
 #ifdef _UNIX
-  if (ArcExt==NULL && (!FileExist(ArcName) || IsDir(GetFileAttr(ArcName))))
-    wcsncatz(ArcName,L".rar",ASIZE(ArcName));
+  // If we want to update an archive without extension, in Windows we can use
+  // "arcname." and it will be treated as "arcname". In Unix "arcname"
+  // and "arcname." are two different names, so we check if "arcname" exists
+  // and do not append ".rar", allowing user to update such archive.
+  if (ExtPos==std::wstring::npos && (!FileExist(ArcName) || IsDir(GetFileAttr(ArcName))))
+    ArcName+=L".rar";
 #else
-  if (ArcExt==NULL)
-    wcsncatz(ArcName,L".rar",ASIZE(ArcName));
+  if (ExtPos==std::wstring::npos)
+    ArcName+=L".rar";
 #endif
   // Treat arcname.part1 as arcname.part1.rar.
-  if (ArcExt!=NULL && wcsnicomp(ArcExt,L".part",5)==0 && IsDigit(ArcExt[5]) &&
-      !FileExist(ArcName))
+  if (ExtPos!=std::wstring::npos && wcsnicomp(&ArcName[ExtPos],L".part",5)==0 &&
+      IsDigit(ArcName[ExtPos+5]) && !FileExist(ArcName))
   {
-    wchar Name[NM];
-    wcsncpyz(Name,ArcName,ASIZE(Name));
-    wcsncatz(Name,L".rar",ASIZE(Name));
+    std::wstring Name=ArcName+L".rar";
     if (FileExist(Name))
-      wcsncpyz(ArcName,Name,ASIZE(ArcName));
+      ArcName=Name;
   }
 
-  if (wcschr(L"AFUMD",*Command)==NULL)
+  if (wcschr(L"AFUMD",Command[0])==NULL && UseStdin.empty())
   {
     if (GenerateArcName)
-      GenerateArchiveName(ArcName,ASIZE(ArcName),GenerateMask,false);
+    {
+      const wchar *Mask=*GenerateMask!=0 ? GenerateMask:DefGenerateMask;
+      GenerateArchiveName(ArcName,Mask,false);
+    }
 
     StringList ArcMasks;
     ArcMasks.AddString(ArcName);
@@ -978,7 +1146,6 @@ void CommandData::ProcessCommand()
     case 'X':
     case 'E':
     case 'T':
-    case 'I':
       {
         CmdExtract Extract(this);
         Extract.DoExtract();
@@ -993,12 +1160,22 @@ void CommandData::ProcessCommand()
       OutHelp(RARX_USERERROR);
 #endif
   }
+
+  // Since messages usually include '\n' in the beginning, we also issue
+  // the final '\n'. It is especially important in Unix, where otherwise
+  // the shell can display the prompt on the same line as the last message.
+  // mprintf is blocked with -idq and if error messages had been displayed
+  // in this mode, we use eprintf to separate them from shell prompt.
+  // If nothing was displayed with -idq, we avoid the excessive empty line.
   if (!BareOutput)
-    mprintf(L"\n");
+    if (MsgStream==MSG_ERRONLY && IsConsoleOutputPresent())
+      eprintf(L"\n");
+    else
+      mprintf(L"\n");
 }
 
 
-void CommandData::AddArcName(const wchar *Name)
+void CommandData::AddArcName(const std::wstring &Name)
 {
   ArcNames.AddString(Name);
 }
@@ -1010,9 +1187,15 @@ bool CommandData::GetArcName(wchar *Name,int MaxSize)
 }
 
 
+bool CommandData::GetArcName(std::wstring &Name)
+{
+  return ArcNames.GetString(Name);
+}
+
+
 bool CommandData::IsSwitch(int Ch)
 {
-#if defined(_WIN_ALL) || defined(_EMX)
+#ifdef _WIN_ALL
   return Ch=='-' || Ch=='/';
 #else
   return Ch=='-';
@@ -1021,7 +1204,7 @@ bool CommandData::IsSwitch(int Ch)
 
 
 #ifndef SFX_MODULE
-uint CommandData::GetExclAttr(const wchar *Str)
+uint CommandData::GetExclAttr(const wchar *Str,bool &Dir)
 {
   if (IsDigit(*Str))
     return wcstol(Str,NULL,0);
@@ -1031,14 +1214,14 @@ uint CommandData::GetExclAttr(const wchar *Str)
   {
     switch(toupperw(*Str))
     {
-#ifdef _UNIX
       case 'D':
-        Attr|=S_IFDIR;
+        Dir=true;
         break;
+#ifdef _UNIX
       case 'V':
         Attr|=S_IFCHR;
         break;
-#elif defined(_WIN_ALL) || defined(_EMX)
+#elif defined(_WIN_ALL)
       case 'R':
         Attr|=0x1;
         break;
@@ -1048,9 +1231,6 @@ uint CommandData::GetExclAttr(const wchar *Str)
       case 'S':
         Attr|=0x4;
         break;
-      case 'D':
-        Attr|=0x10;
-        break;
       case 'A':
         Attr|=0x20;
         break;
@@ -1065,21 +1245,6 @@ uint CommandData::GetExclAttr(const wchar *Str)
 
 
 
-#ifndef SFX_MODULE
-bool CommandData::CheckWinSize()
-{
-  // Define 0x100000000 as macro to avoid troubles with older compilers.
-  const uint64 MaxDictSize=INT32TO64(1,0);
-  // Limit the dictionary size to 4 GB.
-  for (uint64 I=0x10000;I<=MaxDictSize;I*=2)
-    if (WinSize==I)
-      return true;
-  WinSize=0x400000;
-  return false;
-}
-#endif
-
-
 #ifndef SFX_MODULE
 void CommandData::ReportWrongSwitches(RARFORMAT Format)
 {
@@ -1109,3 +1274,57 @@ void CommandData::ReportWrongSwitches(RARFORMAT Format)
   }
 }
 #endif
+
+
+// Get size for string with optional trailing modifiers like "100m".
+int64 CommandData::GetModSize(const wchar *S,uint DefMultiplier)
+{
+  int64 Size=0,FloatingDivider=0;
+  for (uint I=0;S[I]!=0;I++)
+    if (IsDigit(S[I]))
+    {
+      Size=Size*10+S[I]-'0';
+      FloatingDivider*=10;
+    }
+    else
+      if (S[I]=='.')
+        FloatingDivider=1;
+
+  if (*S!=0)
+  {
+    const wchar *ModList=L"bBkKmMgGtT";
+    const wchar *Mod=wcschr(ModList,S[wcslen(S)-1]);
+    if (Mod==nullptr)
+      Size*=DefMultiplier;
+    else
+      for (ptrdiff_t I=2;I<=Mod-ModList;I+=2)
+        Size*=((Mod-ModList)&1)!=0 ? 1000:1024;
+  }
+  if (FloatingDivider!=0)
+    Size/=FloatingDivider;
+  return Size;
+}
+
+
+// Treat the list like rar;zip as *.rar;*.zip for -ms and similar switches.
+void CommandData::GetBriefMaskList(const std::wstring &Masks,StringList &Args)
+{
+  size_t Pos=0;
+  while (Pos<Masks.size())
+  {
+    if (Masks[Pos]=='.')
+      Pos++;
+    size_t EndPos=Masks.find(';',Pos);
+    std::wstring Mask=Masks.substr(Pos,EndPos==std::wstring::npos ? EndPos:EndPos-Pos);
+    if (Mask.find_first_of(L"*?.")==std::wstring::npos)
+      Mask.insert(0,L"*.");
+    Args.AddString(Mask);
+    if (EndPos==std::wstring::npos)
+      break;
+    Pos=EndPos+1;
+  }
+}
+
+
+
+
